from smart_open import smart_open_lib

from .common import PurePath, _SkeletonPath
from .uri import _UriFlavour


class _HttpFlavour(_UriFlavour):
    def splitroot(self, part, sep=None):
        sep = self.sep if sep is None else sep
        drv, root, part = super().splitroot(part, sep)

        if not any((drv == "" or drv.startswith(p) for p in ["http", "https"])):
            raise ValueError(f"http and https are only supported. but {drv} was given.")

        return drv, root, part


_http_flavour = _HttpFlavour()


class PureHttpPath(PurePath):
    _flavour = _http_flavour
    __slots__ = ()


class HttpPath(_SkeletonPath, PureHttpPath):
    __slots__ = ()

    def open(self, *args, **kwargs):
        return smart_open_lib.open(str(self), *args, **kwargs)

    def exists(self):
        from requests.exceptions import HTTPError

        try:
            self.open()
        except HTTPError:
            return False
        return True



from .common import Path, PurePath
from .http import HttpPath, PureHttpPath
from .posix import PosixPath, PurePosixPath
from .s3 import PureS3Path, S3Path
from .uri import PureUriPath
from .windows import PureWindowsPath, WindowsPath



import pathlib

from .common import Path, PurePath


class PurePosixPath(PurePath, pathlib.PurePosixPath):
    __slots__ = ()


class PosixPath(Path, PurePosixPath, pathlib.PosixPath):
    __slots__ = ()



import fnmatch
import pathlib
import posixpath
import re
from os import fsencode
from urllib.parse import quote_from_bytes, urlparse

from .common import Path, PurePath


class _UriFlavour(pathlib._Flavour):
    sep = "/"
    altsep = ""
    has_drv = True
    pathmod = type("fakepath", (), {"normpath": lambda _, x: x})()

    is_supported = True

    def splitroot(self, part, sep=sep):
        url = urlparse(part)
        scheme = url.scheme
        netloc = url.netloc
        path = url.path

        if scheme == "" and netloc != "":
            path = f"{sep}{netloc}{sep}{path}".rstrip(sep)
            netloc = ""

        drv = f"{scheme}://{netloc}" if scheme != "" else ""
        part = path.lstrip("/")
        root = sep if drv != "" or path != part else ""
        return drv, root, part

    def casefold(self, s):
        return s

    def casefold_parts(self, parts):
        return parts

    def compile_pattern(self, pattern):
        return re.compile(fnmatch.translate(pattern)).fullmatch

    def resolve(self, path, strict=False):
        norm_path = posixpath.normpath(self.join(path.parts[1:]))
        return f"{path.drive}{path.root}{norm_path}"

    def is_reserved(self, parts):
        return False

    def make_uri(self, path):
        bpath = bytes(fsencode(self.join(path.parts[1:])))
        return f"{path.drive}{path.root}{quote_from_bytes(bpath)}"

    def gethomedir(self, username):
        raise NotImplementedError("gethomedir() not available on this system")


_uri_flavour = _UriFlavour()


class PureUriPath(PurePath):
    _flavour = _uri_flavour
    __slots__ = ()



import boto3
from botocore.exceptions import ClientError
from smart_open import smart_open_lib

from .common import PurePath, _SkeletonPath
from .uri import _UriFlavour


def _to_file_key(key) -> str:
    return key.rstrip("/")


def _to_dir_key(key) -> str:
    return key if key.endswith("/") else f"{key}/"


class _S3Flavour(_UriFlavour):
    def splitroot(self, part, sep=None):
        sep = self.sep if sep is None else sep
        drv, root, part = super().splitroot(part, sep)

        if drv != "" and not drv.startswith("s3://"):
            raise ValueError(f"http and https are only supported. but {drv} was given.")

        return drv, root, part


_s3_flavour = _S3Flavour()


class PureS3Path(PurePath):
    _flavour = _s3_flavour
    __slots__ = ()

    @property
    def bucket(self):
        if self.anchor == "":
            return ""
        return self.anchor[5:-1]

    @property
    def key(self):
        return self._flavour.sep.join(self.parts[1:])


class S3Path(_SkeletonPath, PureS3Path):
    __slots__ = ()
    _client = boto3.client("s3")  # low-level client is thread safe

    def open(self, *args, **kwargs):
        return smart_open_lib.open(str(self), *args, **kwargs)

    def touch(self, mode=0x666, exist_ok=True):
        if not self._exists(_to_dir_key(self.parent.key)):
            raise FileNotFoundError

        if exist_ok:
            try:
                source = f"{self.bucket}/{self.key}"
                self._client.copy_object(
                    CopySource=source, Bucket=self.bucket, Key=self.key
                )
            except ClientError:
                pass
            else:
                return
        if not exist_ok and self.exists():
            raise FileExistsError
        self._client.put_object(Bucket=self.bucket, Key=self.key)

    def exists(self):
        file_key = _to_file_key(self.key)
        dir_key = _to_dir_key(self.key)
        return self._exists(file_key) or self._exists(dir_key)

    def _mkdir(self, mode=0x777):
        if not self._exists(_to_dir_key(self.parent.key)):
            raise FileNotFoundError

        if self.exists():
            raise OSError
        self._client.put_object(Bucket=self.bucket, Key=_to_dir_key(self.key))

    def iterdir(self):
        def _get_child(content, base):
            return content["Key"].lstrip(base).lstrip("/").split("/")[0]

        continuation_token = ""
        while True:
            cur = self._client.list_objects_v2(
                Bucket=self.bucket,
                Prefix=self.key,
                ContinuationToken=continuation_token,
            )
            names = (_get_child(c, cur["Prefix"]) for c in cur.get("Contents", []))
            for name in names:
                if name in {".", "..", ""}:
                    continue
                yield S3Path(f"{self.anchor}{self.key}/{name}")
            if not cur["IsTruncated"]:
                break
            continuation_token = cur["NextContinuationToken"]

    def is_dir(self):
        try:
            dir_key = _to_dir_key(self.key)
            res = self._client.head_object(Bucket=self.bucket, Key=dir_key)
            return res["ContentLength"] == 0
        except ClientError:
            return False

    def _exists(self, key) -> bool:
        if key == "" or key == "/":
            return True

        try:
            self._client.head_object(Bucket=self.bucket, Key=key)
        except ClientError as e:
            return False
        return True



import os
import pathlib


class PurePath(pathlib.PurePath):
    def __new__(cls, *args):
        cls = cls.get_concrete_cls(*args)
        return cls._from_parts(args)

    @classmethod
    def get_concrete_cls(cls, *args):
        if cls is not PurePath:
            return cls

        if 0 < len(args) and (
            args[0].startswith("http://") or args[0].startswith("https://")
        ):
            from .http import PureHttpPath

            return PureHttpPath

        if 0 < len(args) and args[0].startswith("s3://"):
            from .s3 import PureS3Path

            return PureS3Path

        if os.name == "nt":
            from .windows import PureWindowsPath

            return PureWindowsPath

        from .posix import PurePosixPath

        return PurePosixPath


class Path(PurePath, pathlib.Path):
    def __new__(cls, *args, **kwargs):
        cls = cls.get_concrete_cls(*args)
        self = cls._from_parts(args, init=False)
        if not self._flavour.is_supported:
            raise NotImplementedError(
                "cannot instantiate %r on your system" % (cls.__name__,)
            )
        self._init()
        return self

    @classmethod
    def get_concrete_cls(cls, *args):
        if cls is not Path:
            return cls

        if 0 < len(args) and (
            args[0].startswith("http://") or args[0].startswith("https://")
        ):
            from .http import HttpPath

            return HttpPath

        if 0 < len(args) and args[0].startswith("s3://"):
            from .s3 import S3Path

            return S3Path

        if os.name == "nt":
            from .windows import WindowsPath

            return WindowsPath

        from .posix import PosixPath

        return PosixPath


class _SkeletonAccessor(pathlib._Accessor):
    @staticmethod
    def utime(*args, **kwargs):
        raise NotImplementedError("utime() not available on this system")

    @staticmethod
    def open(*args, **kwargs):
        raise NotImplementedError("open() not available on this system")

    @staticmethod
    def stat(*args, **kwargs):
        raise NotImplementedError("stat() not available on this system")

    @staticmethod
    def lstat(*args, **kwargs):
        raise NotImplementedError("lstat() not available on this system")

    @staticmethod
    def listdir(*args, **kwargs):
        raise NotImplementedError("listdir() not available on this system")

    @staticmethod
    def scandir(*args, **kwargs):
        raise NotImplementedError("scandir() not available on this system")

    @staticmethod
    def chmod(*args, **kwargs):
        raise NotImplementedError("chmod() not available on this system")

    @staticmethod
    def lchmod(*args, **kwargs):
        raise NotImplementedError("lchmod() not available on this system")

    @staticmethod
    def mkdir(*args, **kwargs):
        raise NotImplementedError("mkdir() not available on this system")

    @staticmethod
    def unlink(*args, **kwargs):
        raise NotImplementedError("unlink() not available on this system")

    @staticmethod
    def link_to(*args, **kwargs):
        raise NotImplementedError("link_to() not available on this system")

    @staticmethod
    def rmdir(*args, **kwargs):
        raise NotImplementedError("rmdir() not available on this system")

    @staticmethod
    def rename(*args, **kwargs):
        raise NotImplementedError("rename() not available on this system")

    @staticmethod
    def replace(*args, **kwargs):
        raise NotImplementedError("replace() not available on this system")

    @staticmethod
    def symlink(*args, **kwargs):
        raise NotImplementedError("symlink() not available on this system")

    @staticmethod
    def readlink(*args, **kwargs):
        raise NotImplementedError("readlink() is unsupported on this system")

    @staticmethod
    def owner(*args, **kwargs):
        raise NotImplementedError("owner() is unsupported on this system")

    @staticmethod
    def group(*args, **kwargs):
        raise NotImplementedError("group() is unsupported on this system")


_skeleton_accessor = _SkeletonAccessor()


class _SkeletonPath(Path):
    __slots__ = ()

    def _init(self, template=None):
        super()._init(template)
        self._accessor = _skeleton_accessor

    @classmethod
    def home(cls):
        raise NotImplementedError("home() is not supported.")

    def samefile(self, other_path):
        raise NotImplementedError("samefile() is not supported.")

    def iterdir(self):
        raise NotImplementedError("iterdir() is not supported.")

    def glob(self, pattern):
        raise NotImplementedError("glob() is not supported")

    def rglob(self, pattern):
        raise NotImplementedError("rglob() is not supported")

    def absolute(self):
        raise NotImplementedError("absolute() is not supported")

    def stat(self):
        raise NotImplementedError("stat() is not supported")

    def owner(self):
        raise NotImplementedError("owner() is not supported")

    def group(self):
        raise NotImplementedError("group() is not supported")

    def open(self, mode="r", buffering=-1, encoding=None, errors=None, newline=None):
        raise NotImplementedError("open() is not supported")

    def readlink(self):
        raise NotImplementedError("readlink() is not supported")

    def touch(self, mode=0o666, exist_ok=True):
        raise NotImplementedError("touch() is not supported")

    def _mkdir(self, mode):
        raise NotImplementedError("_mkdir() is not supported")

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        try:
            self._mkdir(mode)
        except FileNotFoundError:
            if not parents or self.parent == self:
                raise
            self.parent.mkdir(parents=True, exist_ok=True)
            self.mkdir(mode, parents=False, exist_ok=exist_ok)
        except OSError:
            if not exist_ok or not self.is_dir():
                raise

    def chmod(self, mode):
        raise NotImplementedError("chmod() is not supported")

    def lchmod(self, mode):
        raise NotImplementedError("lchmod() is not supported")

    def unlink(self, missing_ok=False):
        raise NotImplementedError("unlink() is not supported")

    def rmdir(self):
        raise NotImplementedError("rmdir() is not supported")

    def lstat(self):
        raise NotImplementedError("lstat() is not supported")

    def link_to(self, target):
        raise NotImplementedError("link_to() is not supported")

    def rename(self, target):
        raise NotImplementedError("rename() is not supported")

    def replace(self, target):
        raise NotImplementedError("replace() is not supported")

    def symlink_to(self, target, target_is_directory=False):
        raise NotImplementedError("symlink_to() is not supported")

    def exists(self):
        raise NotImplementedError("exists() is not supported")

    def is_dir(self):
        raise NotImplementedError("is_dir() is not supported")

    def is_file(self):
        raise NotImplementedError("is_file() is not supported")

    def is_mount(self):
        return False

    def is_symlink(self):
        return False

    def is_block_device(self):
        return False

    def is_char_device(self):
        return False

    def is_fifo(self):
        return False

    def is_socket(self):
        return False

    def expanduser(self):
        raise NotImplementedError("expanduser() is not supported")



import pathlib

from .common import Path, PurePath


class PureWindowsPath(PurePath, pathlib.PureWindowsPath):
    __slots__ = ()


class WindowsPath(Path, PureWindowsPath, pathlib.WindowsPath):
    __slots__ = ()
